#-----------------------------------------------------------------------------
#
#   PyWeek 10 - Sprites
#
#-----------------------------------------------------------------------------

from math import hypot, sqrt
from numpy import minimum, maximum
from albow.resource import get_sound
from vectors import Point, Vector, zero_vector
from texture import Texture, TextureArray


FACE_LEFT = 0
FACE_RIGHT = 1
FACE_FRONT = 2

ACT_STAND = 0
ACT_WALK = 1
ACT_SONIC = 2

DIR_LEFT = 0
DIR_RIGHT = 1
DIR_DOWN = 2
DIR_UP = 3

LOW_JUMP = 1
HIGH_JUMP = 2

facing_vectors = [Vector(-1, 0), Vector(1, 0), Vector(0, 0)]
facing_signs = [-1, 1, 0]

walk_speed = 160  # pixels per second
standing_jump_speed = 160
running_jump_speed = 320 #240
vertical_jump_speeds = {LOW_JUMP: 360, HIGH_JUMP: 560}
#standing_jump_velocity = Vector(200, 400) # pixels per second
#running_jump_velocity = Vector(320, 560) # pixels per second
#gravity = -2560    # pixels per square second
gravity = -1280    # pixels per square second

max_movement_step = 8 # pixels


class Sprite(object):
	#  An object moving around on the grid
	#
	#  position          Point       Position of foot in world coordinates
	#  velocity          Vector      Pixels per second
	#  texture           Texture
	#  collision_extent  (w/2, h)    Half-width and height of collision rect
	#  scale             float
	
	#foot_clearance_cells = 2
	max_step_up = 0              #  Number of pixels that the sprite can step up
	step_up_speed = 8            #  Pixels per frame

	facing = FACE_FRONT
	target_step_up = 0
	stepped_up = 0
	delta_step_up = 0            #  1 to step up, -1 to step down, 0 otherwise
	anchor = None
	stuck = False
	
	def __init__(self, texture, scale = 1.0, coords = (0, 0)):
		self.position = Point(*coords)
		self.velocity = zero_vector
		self.texture = texture
		self.scale = scale
		if texture:
			w, h = texture.size
			self.collision_extent = [scale * w / 2, scale * h]

	def facing_vector(self):
		return facing_vectors[self.facing]
	
	def facing_sign(self):
		return facing_signs[self.facing]

	def is_stepping_up(self):
		return self.anchor and self.stepped_up <> self.target_step_up
	
	def is_stepped_up(self):
		return self.stepped_up > 0
	
	def available_step_up(self):
		return self.max_step_up - self.stepped_up
	
	def begin_stepping_up(self, distance):
		#print "Sprite.begin_stepping_up:", distance ###
		self.target_step_up = distance
		self.delta_step_up = 1
	
	def finish_stepping_up(self):
		#print "Sprite.finish_stepping_up" ###
		self.position = self.stepped_up_position()
		self.stepped_up = 0
		self.target_step_up = 0
		self.delta_step_up = 0
	
	def cancel_step_up(self):
		if self.is_stepping_up():
			#print "Sprite.cancel_step_up" ###
			self.delta_step_up = -1
	
	def abort_step_up(self):
		self.stepped_up = 0
		self.target_step_up = 0
		self.delta_step_up = 0
	
	def step_up_frame(self):
		#print "Sprite.step_up_frame" ###
		dy = self.delta_step_up
		y = self.stepped_up + dy * self.step_up_speed
		self.stepped_up = max(0, min(y, self.target_step_up))
		#print "...stepped_up is now",self.stepped_up  ###
		if dy < 0 and self.stepped_up <= 0:
			self.abort_step_up()
	
	def stepped_up_position(self):
		return self.position + Vector(0, self.stepped_up)

	def update_physics(self, game, dt):
		if self.is_stepping_up():
			if self.anchor:
				self.step_up_frame()
			else:
				self.abort_step_up()
		else:
			anchor = self.anchor
			if anchor:
				grid = game.state.grid
				self.position = grid.anchor_position(anchor)
				self.velocity = zero_vector
				self.stuck = False
			else:
				self.abort_step_up()
			self.update_velocity(game, dt)
			dp = self.velocity * dt
			#print "Sprite.update_physics: moving by", dp ###
			moved = self.move(game, dp)
			if moved:
				#print "...moved" ###
				self.update_animation(game)
				self.velocity = self.velocity + Vector(0.0, gravity * dt)
				self.stuck = False
			else:
				#print "...blocked" ###
				if not dp[0] and dp[1] < 0.0:
					self.stuck = True
				#self.velocity = zero_vector
	
	def move(self, game, vector):
		x, y = vector
		if not (x or y):
			return True
		move_sprite = game.state.move_sprite
		if not y:
			return move_sprite(self, Vector(x, 0.0))
		d = abs(y)
		s = max_movement_step
		blocked = False
		while d > 0.0 and not blocked:
			if d < s:
				s = d
			r = s / d
			dx = r * x
			dy = r * y
			#print "Sprite.move:", dx, dy ###
			if dx and not move_sprite(self, Vector(dx, 0.0)):
				self.velocity = self.velocity * Vector(0.0, 1.0)
				blocked = True
			if dy and not move_sprite(self, Vector(0.0, dy)):
				self.velocity = self.velocity * Vector(1.0, 0.0)
				blocked = True
			x -= dx
			y -= dy
			d -= s
		return not blocked

	def update_velocity(self, game, dt):
		pass
	
	def update_animation(self, game):
		pass
	
	def draw(self, game):
		#pos = sprite.position
		pos = self.stepped_up_position()
		self.draw_at(game, pos)
	
	def draw_at(self, game, pos):
		self.texture.draw_quad_with_foot_at(pos, self.scale)
	
	def sonic(self, game, origin):
		pass

#-----------------------------------------------------------------------------

class Player(Sprite):
	#  facing        FACE_*   Direction in which player is facing
	#  phase         int      Animation phase
	#  sonic_charge  int      Sonic screwdriver battery charge
	#  sonic_dir     int      Sonic screwdriver firing direction
	
	collision_extent = (8, 64)
	max_step_up = 16
	num_phases = 8     # Number of animation phases
	stand_textures = TextureArray(2, "player_stand.png", border = 2)
	walk_textures = TextureArray(8, "player_walk.png", border = 2)
	step_up_textures = TextureArray(4, "player_step_up.png", border = 2)
	sonic_textures = TextureArray(4, "player_sonic.png", border = 2)
	sonic_beam_textures = TextureArray(2, "sonic_beam.png", border = 2)
	sonic_sound = get_sound("sonic_screwdriver.ogg")
	sonic_rect_size = (64, 64)
	sonic_offsets = [(-16, 32), (16, 32), (0, 16), (0, 64)]
	max_sonic_charge = 1000
	
	# state -> index, flip
	stand_map = {FACE_LEFT: (1, 0), FACE_RIGHT: (1, 1), FACE_FRONT: (0, 0)}
	sonic_map = {DIR_LEFT: (0, 0), DIR_RIGHT: (0, 1), DIR_UP: (1, 0), DIR_DOWN: (2, 0)}
	
	def __init__(self):
		Sprite.__init__(self, None)
		self.phase = 2
		self.sonic_charge = self.max_sonic_charge
	
	def update_physics(self, game, dt):
		if game.action == ACT_SONIC and self.sonic_charge > 0:
			self.sonic_charge -= 1
			#print "Player.update_physics: sonic charge remaining =", self.sonic_charge
			sonic_pt = self.stepped_up_position() + self.sonic_offsets[self.sonic_dir]
			game.state.sonic(sonic_pt, self.sonic_rect())
			game.start_sound(self.sonic_sound, loops = -1)
		else:
			game.stop_sound(self.sonic_sound)
		Sprite.update_physics(self, game, dt)
	
	def update_velocity(self, game, dt):
		if game.jump_flag and self.anchor:
			self.velocity = self.jump_vector(game)
			game.jump_flag = False
		else:
			if self.anchor or self.stuck:
				act = game.action
				if act == ACT_WALK:
					self.velocity = self.facing_vector() * walk_speed

	def jump_vector(self, game):
		act = game.action
		f = self.facing_sign()
		if act == ACT_WALK:
			vx = running_jump_speed * f
		else:
			vx = standing_jump_speed * f
		vy = vertical_jump_speeds[game.jump_flag]
		return Vector(vx, vy)

	def update_animation(self, game):
		self.phase = (self.phase + 1) % self.num_phases

	def draw_at(self, game, pos):
		pos = self.stepped_up_position()
		if self.target_step_up > 0:
			dy = self.target_step_up - self.stepped_up
			i = dy // self.step_up_speed
			#print "Player.draw: stepped up by", dy, "phase", i ###
			flip = self.facing == FACE_RIGHT
			self.step_up_textures.draw_quad_with_foot_at(i, pos, flip)
		else:
			act = game.action
			if act == ACT_SONIC:
				i, flip = self.sonic_map[self.sonic_dir]
				self.sonic_textures.draw_quad_with_foot_at(i, pos, flip)
				if self.sonic_charge > 0:
					d = self.sonic_dir
					rect = self.sonic_rect()
					self.sonic_beam_textures.draw_rect(d >> 1, rect, (d & 1) << 1)
			else:
				if self.anchor:
					if act == ACT_STAND:
						i, flip = self.stand_map[self.facing]
						self.stand_textures.draw_quad_with_foot_at(i, pos, flip)
					elif act == ACT_WALK:
						i = self.phase
						flip = self.facing == FACE_RIGHT
						self.walk_textures.draw_quad_with_foot_at(i, pos, flip)
				else:
					flip = self.facing == FACE_RIGHT
					self.walk_textures.draw_quad_with_foot_at(0, pos, flip)
	
	def reset_walk_phase(self):
		self.phase = 2
	
	def sonic_rect(self):
		x, y = self.position
		d = self.sonic_dir
		w, h = self.sonic_rect_size
		dx, dy = self.sonic_offsets[d]
		if d == DIR_LEFT:
			return x + dx - w, y + dy - h/2, x + dx, y + dy + h/2
		elif d == DIR_RIGHT:
			return x + dx, y + dy - h/2, x + dx + w, y + dy + h/2
		elif d == DIR_DOWN:
			return x - w/2, y + dy - h, x + w/2, y + dy
		elif d == DIR_UP:
			return x - w/2, y + dy, x + w/2, y + dy + h

#-----------------------------------------------------------------------------

class Furniture(Sprite):
	pass
